@@ -3,22 +3,48 @@
 from collections import namedtuple
 import math
 import sys
+import warnings
 
 if sys.version_info < (3,):
+    warnings.warn(
+        "Python versions < 3 will not be supported by mercantile 2.0.",
+        DeprecationWarning,
+    )
     from collections import Sequence
 else:
     from collections.abc import Sequence
 
 
-__version__ = '1.1.1'
+__version__ = "1.1.3"
 
 __all__ = [
-    'Bbox', 'LngLat', 'LngLatBbox', 'Tile', 'bounding_tile', 'bounds',
-    'children', 'feature', 'lnglat', 'parent', 'quadkey', 'quadkey_to_tile',
-    'simplify', 'tile', 'tiles', 'ul', 'xy_bounds']
+    "Bbox",
+    "LngLat",
+    "LngLatBbox",
+    "Tile",
+    "bounding_tile",
+    "bounds",
+    "children",
+    "feature",
+    "lnglat",
+    "parent",
+    "quadkey",
+    "quadkey_to_tile",
+    "simplify",
+    "tile",
+    "tiles",
+    "ul",
+    "xy_bounds",
+]
+
+R2D = 180 / math.pi
+RE = 6378137.0
+CE = 2 * math.pi * RE
+EPSILON = 1e-14
+LL_EPSILON = 1e-11
 
 
-Tile = namedtuple('Tile', ['x', 'y', 'z'])
+Tile = namedtuple("Tile", ["x", "y", "z"])
 """An XYZ web mercator tile
 
 Attributes
@@ -28,7 +54,7 @@ x, y, z : int
 """
 
 
-LngLat = namedtuple('LngLat', ['lng', 'lat'])
+LngLat = namedtuple("LngLat", ["lng", "lat"])
 """A longitude and latitude pair
 
 Attributes
@@ -38,7 +64,7 @@ lng, lat : float
 """
 
 
-LngLatBbox = namedtuple('LngLatBbox', ['west', 'south', 'east', 'north'])
+LngLatBbox = namedtuple("LngLatBbox", ["west", "south", "east", "north"])
 """A geographic bounding box
 
 Attributes
@@ -48,7 +74,7 @@ west, south, east, north : float
 """
 
 
-Bbox = namedtuple('Bbox', ['left', 'bottom', 'right', 'top'])
+Bbox = namedtuple("Bbox", ["left", "bottom", "right", "top"])
 """A web mercator bounding box
 
 Attributes
@@ -82,6 +108,10 @@ class TileArgParsingError(MercantileError):
     """Raised when errors occur in parsing a function's tile arg(s)"""
 
 
+class TileError(MercantileError):
+    """Raised when a tile can't be determined"""
+
+
 def _parse_tile_arg(*args):
     """parse the *tile arg of module functions
 
@@ -104,7 +134,9 @@ def _parse_tile_arg(*args):
     if len(args) == 3:
         return Tile(*args)
     else:
-        raise TileArgParsingError("the tile argument may have 1 or 3 values. Note that zoom is a keyword-only argument")
+        raise TileArgParsingError(
+            "the tile argument may have 1 or 3 values. Note that zoom is a keyword-only argument"
+        )
 
 
 def ul(*tile):
@@ -131,9 +163,9 @@ def ul(*tile):
     """
     tile = _parse_tile_arg(*tile)
     xtile, ytile, zoom = tile
-    n = 2.0 ** zoom
-    lon_deg = xtile / n * 360.0 - 180.0
-    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
+    Z2 = math.pow(2, zoom)
+    lon_deg = xtile / Z2 * 360.0 - 180.0
+    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / Z2)))
     lat_deg = math.degrees(lat_rad)
     return LngLat(lon_deg, lat_deg)
 
@@ -150,6 +182,11 @@ def bounds(*tile):
     -------
     LngLatBBox
 
+    Notes
+    -----
+    Epsilon is subtracted from the right limit and added to the bottom
+    limit.
+
     """
     tile = _parse_tile_arg(*tile)
     xtile, ytile, zoom = tile
@@ -189,14 +226,16 @@ def xy(lng, lat, truncate=False):
     """
     if truncate:
         lng, lat = truncate_lnglat(lng, lat)
-    x = 6378137.0 * math.radians(lng)
+
+    x = RE * math.radians(lng)
+
     if lat <= -90:
-        y = float('-inf')
+        y = float("-inf")
     elif lat >= 90:
-        y = float('inf')
+        y = float("inf")
     else:
-        y = 6378137.0 * math.log(
-            math.tan((math.pi * 0.25) + (0.5 * math.radians(lat))))
+        y = RE * math.log(math.tan((math.pi * 0.25) + (0.5 * math.radians(lat))))
+
     return x, y
 
 
@@ -215,11 +254,10 @@ def lnglat(x, y, truncate=False):
     LngLat
 
     """
-    R2D = 180 / math.pi
-    A = 6378137.0
     lng, lat = (
-        x * R2D / A,
-        ((math.pi * 0.5) - 2.0 * math.atan(math.exp(-y / A))) * R2D)
+        x * R2D / RE,
+        ((math.pi * 0.5) - 2.0 * math.atan(math.exp(-y / RE))) * R2D,
+    )
     if truncate:
         lng, lat = truncate_lnglat(lng, lat)
     return LngLat(lng, lat)
@@ -237,14 +275,42 @@ def xy_bounds(*tile):
     -------
     Bbox
 
+    Notes
+    -----
+    Epsilon is subtracted from the right limit and added to the bottom
+    limit.
+
     """
     tile = _parse_tile_arg(*tile)
     xtile, ytile, zoom = tile
-    left, top = xy(*ul(xtile, ytile, zoom))
-    right, bottom = xy(*ul(xtile + 1, ytile + 1, zoom))
+
+    tile_size = CE / math.pow(2, zoom)
+
+    left = xtile * tile_size - CE / 2
+    right = left + tile_size
+
+    top = CE / 2 - ytile * tile_size
+    bottom = top - tile_size
+
     return Bbox(left, bottom, right, top)
 
 
+def _xy(lng, lat, truncate=False):
+
+    if truncate:
+        lng, lat = truncate_lnglat(lng, lat)
+
+    x = lng / 360.0 + 0.5
+    sinlat = math.sin(math.radians(lat))
+
+    try:
+        y = 0.5 - 0.25 * math.log((1.0 + sinlat) / (1.0 - sinlat)) / math.pi
+    except ValueError:
+        raise InvalidLatitudeError("Y can not be computed: lat={!r}".format(lat))
+    else:
+        return x, y
+
+
 def tile(lng, lat, zoom, truncate=False):
     """Get the tile containing a longitude and latitude
 
@@ -262,20 +328,36 @@ def tile(lng, lat, zoom, truncate=False):
     Tile
 
     """
-    if truncate:
-        lng, lat = truncate_lnglat(lng, lat)
-    lat = math.radians(lat)
-    n = 2.0 ** zoom
-    xtile = int(math.floor((lng + 180.0) / 360.0 * n))
+    x, y = _xy(lng, lat, truncate=truncate)
+    Z2 = math.pow(2, zoom)
 
-    try:
-        ytile = int(math.floor((1.0 - math.log(
-            math.tan(lat) + (1.0 / math.cos(lat))) / math.pi) / 2.0 * n))
-    except ValueError:
-        raise InvalidLatitudeError(
-            "Y can not be computed for latitude {} radians".format(lat))
+    if x <= 0:
+        xtile = 0
+    elif x >= 1:
+        xtile = int(Z2 - 1)
     else:
-        return Tile(xtile, ytile, zoom)
+        # Heuristic to find points straddling tiles.
+        xtile_a = math.floor((x - EPSILON) * Z2)
+        xtile_b = math.floor((x + EPSILON) * Z2)
+        if xtile_a != xtile_b:
+            xtile = int(xtile_b)
+        else:
+            xtile = int(xtile_a)
+
+    if y <= 0:
+        ytile = 0
+    elif y >= 1:
+        ytile = int(Z2 - 1)
+    else:
+        # Heuristic to find points straddling tiles.
+        ytile_a = math.floor((y + EPSILON) * Z2)
+        ytile_b = math.floor((y - EPSILON) * Z2)
+        if ytile_a != ytile_b:
+            ytile = int(ytile_a)
+        else:
+            ytile = int(ytile_b)
+
+    return Tile(xtile, ytile, zoom)
 
 
 def quadkey(*tile):
@@ -302,7 +384,7 @@ def quadkey(*tile):
         if ytile & mask:
             digit += 2
         qk.append(str(digit))
-    return ''.join(qk)
+    return "".join(qk)
 
 
 def quadkey_to_tile(qk):
@@ -323,20 +405,24 @@ def quadkey_to_tile(qk):
     xtile, ytile = 0, 0
     for i, digit in enumerate(reversed(qk)):
         mask = 1 << i
-        if digit == '1':
+        if digit == "1":
             xtile = xtile | mask
-        elif digit == '2':
+        elif digit == "2":
             ytile = ytile | mask
-        elif digit == '3':
+        elif digit == "3":
             xtile = xtile | mask
             ytile = ytile | mask
-        elif digit != '0':
+        elif digit != "0":
+            warnings.warn(
+                "QuadKeyError will not derive from ValueError in mercantile 2.0.",
+                DeprecationWarning,
+            )
             raise QuadKeyError("Unexpected quadkey digit: %r", digit)
     return Tile(xtile, ytile, i + 1)
 
 
 def tiles(west, south, east, north, zooms, truncate=False):
-    """Get the tiles intersecting a geographic bounding box
+    """Get the tiles overlapped by a geographic bounding box
 
     Parameters
     ----------
@@ -351,6 +437,11 @@ def tiles(west, south, east, north, zooms, truncate=False):
     ------
     Tile
 
+    Notes
+    -----
+    A small epsilon is used on the south and east parameters so that this
+    function yields exactly one tile when given the bounds of that same tile.
+
     """
     if truncate:
         west, south = truncate_lnglat(west, south)
@@ -363,7 +454,6 @@ def tiles(west, south, east, north, zooms, truncate=False):
         bboxes = [(west, south, east, north)]
 
     for w, s, e, n in bboxes:
-
         # Clamp bounding values.
         w = max(-180.0, w)
         s = max(-85.051129, s)
@@ -374,15 +464,11 @@ def tiles(west, south, east, north, zooms, truncate=False):
             zooms = [zooms]
 
         for z in zooms:
-            ll = tile(w, s, z)
-            ur = tile(e, n, z)
+            ul_tile = tile(w, n, z)
+            lr_tile = tile(e - LL_EPSILON, s + LL_EPSILON, z)
 
-            # Clamp left x and top y at 0.
-            llx = 0 if ll.x < 0 else ll.x
-            ury = 0 if ur.y < 0 else ur.y
-
-            for i in range(llx, min(ur.x + 1, 2 ** z)):
-                for j in range(ury, min(ll.y + 1, 2 ** z)):
+            for i in range(ul_tile.x, lr_tile.x + 1):
+                for j in range(ul_tile.y, lr_tile.y + 1):
                     yield Tile(i, j, z)
 
 
@@ -421,7 +507,8 @@ def parent(*tile, **kwargs):
 
     if zoom is not None and (tile[2] < zoom or zoom != int(zoom)):
         raise InvalidZoomError(
-            "zoom must be an integer and less than that of the input tile")
+            "zoom must be an integer and less than that of the input tile"
+        )
 
     x, y, z = tile
     if x != int(x) or y != int(y) or z != int(z):
@@ -480,7 +567,8 @@ def children(*tile, **kwargs):
 
     if zoom is not None and (ztile > zoom or zoom != int(zoom)):
         raise InvalidZoomError(
-            "zoom must be an integer and greater than that of the input tile")
+            "zoom must be an integer and greater than that of the input tile"
+        )
 
     target_zoom = zoom if zoom is not None else ztile + 1
 
@@ -491,7 +579,7 @@ def children(*tile, **kwargs):
             Tile(xtile * 2, ytile * 2, ztile + 1),
             Tile(xtile * 2 + 1, ytile * 2, ztile + 1),
             Tile(xtile * 2 + 1, ytile * 2 + 1, ztile + 1),
-            Tile(xtile * 2, ytile * 2 + 1, ztile + 1)
+            Tile(xtile * 2, ytile * 2 + 1, ztile + 1),
         ]
     return tiles
 
@@ -571,25 +659,33 @@ def bounding_tile(*bbox, **kwds):
     """
     if len(bbox) == 2:
         bbox += bbox
+
     w, s, e, n = bbox
-    truncate = bool(kwds.get('truncate'))
+
+    truncate = bool(kwds.get("truncate"))
+
     if truncate:
         w, s = truncate_lnglat(w, s)
         e, n = truncate_lnglat(e, n)
-    # Algorithm ported directly from https://github.com/mapbox/tilebelt.
+
+    e = e - LL_EPSILON
+    s = s + LL_EPSILON
 
     try:
-        tmin = tile(w, s, 32, truncate=truncate)
-        tmax = tile(e, n, 32, truncate=truncate)
+        tmin = tile(w, n, 32)
+        tmax = tile(e, s, 32)
     except InvalidLatitudeError:
         return Tile(0, 0, 0)
 
     cell = tmin[:2] + tmax[:2]
     z = _getBboxZoom(*cell)
+
     if z == 0:
         return Tile(0, 0, 0)
+
     x = rshift(cell[0], (32 - z))
     y = rshift(cell[1], (32 - z))
+
     return Tile(x, y, z)
 
 
@@ -597,15 +693,14 @@ def _getBboxZoom(*bbox):
     MAX_ZOOM = 28
     for z in range(0, MAX_ZOOM):
         mask = 1 << (32 - (z + 1))
-        if ((bbox[0] & mask) != (bbox[2] & mask) or
-                (bbox[1] & mask) != (bbox[3] & mask)):
+        if (bbox[0] & mask) != (bbox[2] & mask) or (bbox[1] & mask) != (bbox[3] & mask):
             return z
     return MAX_ZOOM
 
 
 def feature(
-        tile, fid=None, props=None, projected='geographic', buffer=None,
-        precision=None):
+    tile, fid=None, props=None, projected="geographic", buffer=None, precision=None
+):
     """Get the GeoJSON feature corresponding to a tile
 
     Parameters
@@ -631,37 +726,43 @@ def feature(
 
     """
     west, south, east, north = bounds(tile)
-    if projected == 'mercator':
+
+    if projected == "mercator":
         west, south = xy(west, south, truncate=False)
         east, north = xy(east, north, truncate=False)
+
     if buffer:
         west -= buffer
         south -= buffer
         east += buffer
         north += buffer
+
     if precision and precision >= 0:
         west, south, east, north = (
-            round(v, precision) for v in (west, south, east, north))
-    bbox = [
-        min(west, east), min(south, north),
-        max(west, east), max(south, north)]
+            round(v, precision) for v in (west, south, east, north)
+        )
+
+    bbox = [min(west, east), min(south, north), max(west, east), max(south, north)]
     geom = {
-        'type': 'Polygon',
-        'coordinates': [[
-            [west, south],
-            [west, north],
-            [east, north],
-            [east, south],
-            [west, south]]]}
+        "type": "Polygon",
+        "coordinates": [
+            [[west, south], [west, north], [east, north], [east, south], [west, south]]
+        ],
+    }
+
     xyz = str(tile)
     feat = {
-        'type': 'Feature',
-        'bbox': bbox,
-        'id': xyz,
-        'geometry': geom,
-        'properties': {'title': 'XYZ tile %s' % xyz}}
+        "type": "Feature",
+        "bbox": bbox,
+        "id": xyz,
+        "geometry": geom,
+        "properties": {"title": "XYZ tile %s" % xyz},
+    }
+
     if props:
-        feat['properties'].update(props)
-    if fid:
-        feat['id'] = fid
+        feat["properties"].update(props)
+
+    if fid is not None:
+        feat["id"] = fid
+
     return feat
